/*****************************************************************************/
/*
                                   mondesi.c

                          Monitor de Sistema (espaol)

MONDESI (pronounced "mon-deh-sih") is a browser-based VMS system monitor with a
sufficiently small resource footprint as not to impact system performance
(minimal CPU consumption and only one buffered IO per client update).  No, the
name's etiology has little significance at all (a small homage at most).  I was
just looking for a short (and cute) name.

This utility is intended as a general tool for system observation.  It is not a
troubleshooting tool as such, especially when a system is beginning to suffer
from resource exhaustion.  Not only does the data collection script need to
continue processing effectively (see 'SUFFICIENT PRIORITY' below) there is also
a Web server in the data transfer path.

MONDESI should work with all relatively modern browsers (the graphics toolkit -
see below - is the primary limitation); those with the Gecko-Engine - Mozilla,
SeaMonkey, Netscape 6+, Firefox, along with Opera 7+, and MSIE 5+.  The intense
graphing is CPU and memory hungry so I'd not recommend Mozilla on an DEC 3000
(or even a PWS 500 :-)  Of course YMMV.

Uses the system service $GETRMI() available for VMS V7.3 and later.
Note that V7.3-1 (and earlier) do not support some data items.  These are shown
as question marks.

Displayed values have been validated using the interactive MONITOR utility.

As much of the functionality is pushed off to the browser MONDESI relies
heavily on JavaScript contained in MONDESI.JS and WS_JSGRAPHICS.JS (see below).

As MONDESI provides a readily accessable insight into a system (albeit,
basical resource consumption information only) perhaps it would be best
available only when under authorization.

NOTE: MONDESI was essentially a vehicle for experimentation rather than an
application designed and built to address a specific need.  On the other hand
it may be useful to someone.  Enjoy!


MONITOR DISPLAY
---------------
A node is displayed as a text panel on the left and nine bar graphs on the
right.  These display the instantaneous (current) values for the various
resources and are updated dynamically at the interval when the data collector
script detects a change in values.  Each bar graph section contains a number
representing the current value and a number in the far right the maximum value
reached while monitoring.  All values are per-second where applicable (e.g.
BIO, DIO, lock, etc.)

  +-------------------------------------------------------------------------+
  | NODENAME              CPU:[_________________________:_____|___________] |
  | 21-SEP-2008 21:44     MEM:[________________|__________________________] |
  | Digital Personal     PGFL:[___________________________________________] |
  | 1 CPU 766 MB         FLTS:[_____:_|___________________________________] |
  | Alpha V8.3            BIO:[__________________|________________________] |
  | Up: 29 05:50          DIO:[_|_________________________________________] |
  | Processes:25         LOCK:[__________________________________|________] |
  |                      DISK:[__:_____________|__________________________] |
  | [hi][pr][re][ze][st] MSCP:[_______-__________________________________] |
  +-------------------------------------------------------------------------+

CPU: shows total CPU utilisation in green, non-user-mode CPU
(kernel+interrupt+etc.) in brown (always a subset of total) and average CPU
utilisation in yellow.  Where a system has single CPU the graph range is 100%,
with two CPUs the range is 200%, three is 300%, etc.

MEM: displays system physical memory (RAM) in Mbytes.  The value in the far
right is the total.  The current used is show in green.  Used memory is
calculate by subtracting the current free list plus modified list sizes from
the total physical memory.

PGFL: shows page file in Mbytes with current utilisation displayed in green.

FLTS: provides total page faults per second in green with hard page faults as a
subset in brown.  Average page faulting for the monitor period is shown in
yellow.

BIO: displays buffered I/Os per second.  Average is shown in yellow.

DIO: displays direct I/Os per second.  Average is shown in yellow.

LOCK: provides locking activity as a total of ENQs plus DEQs plus waits. 
Average activity is shown in yellow.

DISK: displays total disk I/Os in green with a subset of disk writes in
brown (disk reads being the difference).  Average activity shown in yellow.

MSCP: in a non-clustered system this item will be grayed-out.  Displays total
requests from remote systems in green and a subset showing MSCP write requests
in brown.  Average is displayed in yellow.

[history] button reveals/hides the history line graphs (see below).

[print] suspends updates to allow use of the print dialog

[scale] adjusts the y axis to the current maximum data value

[zero] button can be used to reset the collected maximum and average data.

[stop][start] stops data collection while maintaining the display
              start (re)starts data collection and zeroes display


HISTORY DISPLAY
---------------
Recent data is stored within the browser and used to generate line graphs
showing values against time.  These follow the same colouring schema as the
instantaneous bar graphs.   For example, the CPU consumption line is in green
and the non-user-mode in brown.

The [history] button reveals and hides the history graphs.  The complexity of
rendering these graphics puts a considerable load on the browser system.


PRINTING
--------
Printing a 'page' that is currently building itself (as these displays are
because of the hidden IFRAME receiving the resource data) is problematic on all
browsers tested.   As a workaround the data feed is closed (by removing the
hidden IFRAME) thus stopping the page building.  The JavaScript print is then
requested.  After a second or two the IFRAME is recreated resuming the data
feed and update of the displays.  Bit kludgy, but hey, it's a browser.  Firefox
and MSIE printed OK but Opera provided an unrendered page.  C'est la vie!

The [print] button is highlighted when first clicked.  Display updates (but not
data gathering) are suspended while the [print] button is engaged allowing a
static 'snapshot' of the data to be printed.  The browser print dialog can then
be used or cancelled.  Clicking the [print] button again once the print is
queued (and you should!) resumes display updates.  The [print] button remains
enabled for a maximum of 45 seconds before being automatically reset.


CONFIGURATION
-------------
The configuration page provides a text entry area into which URLs for the nodes
to monitor are entered, one per line.  The 'nodes across window' selector allow
monitor displays to be tiled (if there is enough screen real estate).  Using
this approach widely 'separated' hosts can be monitored on a single page.  The
URL generated by the configuration dialog can be saved as a bookmark, entered
into an HTML page link, etc., for ease of subsequent access.


SUFFICIENT PRIORITY
-------------------
For a monitor such as this to be effective on a busy system it needs to have a
bit of an edge!  If INSTALLed with ALTPRI it will bump it's process priority up
to 6 when starting and reduce it back to previous when exiting (just in case
it's WASD and a persistent scripting process :-) 

  $ INSTALL REPLACE <location>MONDESI.EXE /PRIV=ALTPRI

Don't forget to reINSTALL any subsequently built or deployed executable.


WEB "PUSH"
----------
MONDESI uses JavaScript and a Web 'push' technology (such as it is) to return
data (not images!) from a Web server scripting process to the client monitor
page where it is displayed as text or rendered into (crude but effective) bar
graphs.  Perhaps it would be better described as "HTTP streaming" or as "long
polling" but it *definitely* does not need to poll for each update.  The "push"
will last as long as the Web server and intermediate network connectivity
allows.

A parent page provides the JavaScript and HTML infrastructure to support
multiple monitored nodes each contained in an indepdendent, visible IFRAME,
which in turn used a dynamically created (i.e. JavaScript), hidden IFRAME to
execute the data collection URL (script).  This executes continuously,
periodically collecting data and outputting changed items into the response
stream.  Each response record contains a JavaScript tag element which contains
calls to the required data processing and display functions in the visible
IFRAME to process (only) each datum that needs updating.

If a data collection process (script) should exit for whatever reason (timeout,
network interruption, etc.) MONDESI should restart data collection
automatically.  If the monitored system become unavailable for any reason
(restart, network interruption, etc.) the monitor highlights the node
information with red after ninety seconds.  Elementary but suitable to the
desired outcome.


INTERFRAME COMMUNICATION (IFC)
------------------------------
The node(s) displayed in the parent page are not necessarily the same as the
parent page (indeed, multiple nodes could not all be) and of course these have
differing host names and/or IP addresses.  This incurs browser cross-domain
restrictions with communication between the node display HTML (i)frame and the
primary, top-level page, specifically - when needing to open/close the history
graph portion of the display associated with each node.

Two approaches for providing this IFC are used. 
Both are well documented on the Web.

1) The more elegant approach uses the modern, architected, HTML5 postMessage()
communication mechanism.  MONDESI checks for and uses this if available in the
browser.  It sends a message to the top-level page which invokes a function to
adjust the node display height.  Currently (at initial MONDESI development)
Firefox 3 (specifically Gecko 1.9), Opera 9.5 and MSIE 8 (at least) support it.

2) If postMessage() is unavailable MONDESI falls back to using an IFC *kludge*
involving the dynamic creation of an IFRAME within the node monitor IFRAME and
using this to load a URL for the served file BCHAN.HTML which contains a call
to a JavaScript function in the top-level page (to adjust the height).  The
data is contained in the hash portion of the URL and never appears outside the
browser.  This should be usable with most older, mainstream browsers.


GRAPHING
--------
This application uses a third-party JavaScript graphics toolkit to generate
both the bar and line graphs.

  Copyright (c) 2002-2004 Walter Zorn. All rights reserved.

  Performance optimizations for Internet Explorer
  by Thomas Frank and John Holdsworth.
  fillPolygon method implemented by Matthieu Haller.

  LICENSE: LGPL

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

Documented at

  http://www.walterzorn.com/jsgraphics/jsgraphics_e.htm

Many thanks to Walter for making this useful tool freely available.  This
toolkit works extraordinarily well considering it uses no graphics capability
in the browser!!  This is a non-SVG, non-VML, DOM kludge (not meaning this as a
pejorative).


BUILD DETAILS
-------------
See BUILD_MONDESI.COM procedure.

  $ @BUILD_MONDESI [LINK]


APACHE INSTALLATION
-------------------
Copy the executable to the APACHE$ROOT:[CGI-BIN] directory.

Map the JavaScript resources into web-space using (assumes the application is
located under APACHE$COMMON:[000000]):

  Alias /mondesi/-/ "/apache$common/mondesi/"


OSU ACCOMODATIONS
-----------------
OSU requires the use of the MONDESI.COM DCL wrapper to invoke the OSU CGI
environment utility before executing MONDESI.  The wrapper and the MONDESI.EXE
executable both should be placed in WWW_ROOT:[BIN].  Rather than introduce the
bulk of WASD CGILIB this deliberately slimline monitor utility checks for an
indication of the OSU environment and overtly handles the indiosyncracies.  
MONDESI does not reopen <stdout> in binary mode.  OSU requires two CGI response
headers (MONDESI supplies "Status:" and "Content-Type:").  OSU also dislikes a
response header with <CR><LF> carriage-control, so MONDESI uses only an <LF>.

Map the JavaScript resources into web-space using (assumes the application is
located under WWW_ROOT:[000000]):

  pass /mondesi/-/* /www_root/mondesi/*


WASD INSTALLATION
-----------------
Copy the executable to the CGI_EXE: directory.

Map the JavaScript resources into web-space using (assumes the application is
located in the source tree):

  # HTTPD$MAP
  pass /mondesi/-/* /ht_root/src/mondesi/*


COPYRIGHT
---------
Copyright (C) 2008 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY (update SOFTWAREVN as well!)
---------------
13-DEC-2008  MGD  v1.0.1, CollectRMI() allow for 7.3-1 and earlier (sigh!)
                          Thanks Graham Burley for suggesting sys$gl_*jobcnt
29-NOV-2008  MGD  v1.0.0, initial
*/
/*****************************************************************************/

#define SOFTWAREVN "1.0.1"
#define SOFTWARENM "MONDESI"
#ifdef __ALPHA
   char SoftwareID [] = SOFTWARENM " AXP-" SOFTWAREVN;
#endif
#ifdef __ia64
   char SoftwareID [] = SOFTWARENM " IA64-" SOFTWAREVN;
#endif
#ifdef __VAX
   char SoftwareID [] = SOFTWARENM " VAX-" SOFTWAREVN;
#endif

/* standard C header files */
#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* VMS-related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdtdef.h>
#include <lnmdef.h>
#include <psldef.h>
#include <rmidef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* required (to compile) for VMS V7.3-1 and earlier */
#ifndef RMI$_CPUCOMPAT
#define RMI$_CPUCOMPAT 21193
#define RMI$_CPUINTSTK 21194
#define RMI$_CPUMPSYNCH 21195
#define RMI$_CPUKERNEL 21196
#define RMI$_CPUEXEC 21197
#define RMI$_CPUSUPER 21198
#define RMI$_CPUUSER 21199
#define RMI$_CPUIDLE 21200
#define RMI$_PROCBATCNT 21202
#define RMI$_PROCINTCNT 21203
#define RMI$_PROCNETCNT 21204
#define RMI$_MSCP_EVERYTHING 4265
#endif

/* required for VMS V7.3-1 and earlier */
#ifdef  __VAX
globalref sys$gw_ijobcnt;
globalref sys$gw_bjobcnt;
globalref sys$gw_njobcnt;
#else
globalref sys$gl_ijobcnt;
globalref sys$gl_bjobcnt;
globalref sys$gl_njobcnt;
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

#define BOOL int
#define TRUE 1
#define FALSE 0
 
#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))
 
#define ALERT_SECONDS 90
#define AMBIT_PRIORITY 6
#define REFRESH_INTERVAL 2

#define NODE_HEIGHT 148
#define NODE_WIDTH 952

/* needs to be a multiple of nine */
#define BAR_HEIGHT      14
#define BAR_WIDTH       (HISTORY_WIDTH+2 + HISTORY_WIDTH+2 + 10)

#define HISTORY_HEIGHT  100
#define HISTORY_WIDTH   300

/*******************/
/* data structures */
/*******************/

#ifndef __VAX
#   pragma member_alignment __save
#   pragma nomember_alignment
#endif

typedef struct STRUCT_IOSBLK
{
   unsigned short  Status;
   unsigned short  Count;
   unsigned long  Unused;
}
IOSBLK;

#ifndef __VAX
#   pragma member_alignment __restore
#endif

typedef struct STRUCT_SYSTEM_DATA
{
   unsigned long  ActiveCpuCount,
                  AvailCpuCount,
                  BufferedIO,
                  BufferedIORaw,
                  ClusterMember,
                  ClusterEVotes,
                  ClusterNodes,
                  ClusterQuorum,
                  ClusterVotes,
                  CpuCount,
                  DirectIO,
                  DirectIORaw,
                  FcpRead,
                  FcpReadRaw,
                  FcpWrite,
                  FcpWriteRaw,
                  FreeList,
                  HardPageFaults,
                  LockActivity,
                  LockActivityRaw,
                  LockWait,
                  MemInUse,
                  MemInUseMax,
                  MemSize,
                  ModList,
                  MscpRead,
                  MscpReadRaw,
                  MscpWrite,
                  MscpWriteRaw,
                  PrevHardPageFaults,
                  PageFaults,
                  PageFaultsHard,
                  PageFaultsRaw,
                  PageFaultsSoft,
                  PageFileFree,
                  PageFilePage,
                  PageReadIO,
                  PageSize,
                  PageWriteIO,
                  ProcessCount,
                  ProcessBatCount,
                  ProcessIntCount,
                  ProcessNetCount,
                  PrevBufferedIO,
                  PrevDirectIO,
                  PrevFcpRead,
                  PrevFcpWrite,
                  PrevLockActivity,
                  PrevLockWait,
                  PrevMscpRead,
                  PrevMscpWrite,
                  PrevPageFaults,
                  PrevPageFileFree,
                  PrevPageReadIO,
                  PrevPageWriteIO,
                  PrevMscpIO,
                  Processes;
 
   char  ArchName [15+1],
         HwName [60+1],
         NodeName [15+1],
         VmsVersion [8+1];

   unsigned long  BootBinTime [2],
                  PrevBinTime [2],
                  CpuTicksBusy [2],
                  CpuTicksUser [2],
                  CpuTicksRaw [2],
                  PrevCpuTicksBusy [2],
                  PrevCpuTicksUser [2];

   int  CpuBusyPercent,
        CpuUserPercent;
}
SYSTEM_DATA;

/******************/
/* global storage */
/******************/

BOOL  Debug,
      ApacheHttpd,
      OsuHttpd;

int  PrevPrio;

unsigned long  CurrentBinTime [2];
unsigned short  CurrentNumTime [7];

char  *CgiHttpRefererPtr,
      *CgiScriptNamePtr,
      *CgiQueryStringPtr;

char  RefererHost [128];

/**************/
/* prototypes */
/**************/

int CollectRMI (SYSTEM_DATA*);
int CollectSyiStatic (SYSTEM_DATA*);
int CollectSyiDynamic (SYSTEM_DATA*);
char* GetCgiVar (char*);
char* GetCgiVarNull (char*);
void MonDeSiData (SYSTEM_DATA*, int);
void MonDeSiForm (char*, int, int);
void MonDeSiNode (SYSTEM_DATA*, char*, int);
void MonDeSiPage (char*, int, int);
void ReturnPrio ();

/*****************************************************************************/
/*
*/

int main ()

{
   int  status,
        NodesAcross = 1,
        RefreshInterval = REFRESH_INTERVAL;
   char  *cptr, *sptr, *zptr,
         *NodeList;
   char  IframeId [64] = "";
   SYSTEM_DATA  SystemData;

   /*********/
   /* begin */
   /*********/

   /* ambit attempt to bump the process priority (if INSTALLed with ALTPRI) */
   atexit (&ReturnPrio);
   sys$setpri (0, 0, AMBIT_PRIORITY, &PrevPrio, 0, 0);

   if (getenv ("MONDESI$DBUG")) Debug = 1;

   if (Debug)
      fprintf (stdout,
"Status: 200\n\
Content-Type: text/plain\n\
Script-Control: X-content-encoding-gzip=0\n\
\r\n");

   ApacheHttpd = (getenv ("APACHE$SHARED_SOCKET") != NULL ||
                  getenv ("APACHE$PARENT_PID") != NULL ||
                  getenv ("APACHE$SERVER_TAG") != NULL);
   OsuHttpd = (getenv("WWWEXEC_RUNDOWN_STRING") != NULL);

   if (!Debug && !OsuHttpd)
      if (!(stdout = freopen ("SYS$OUTPUT:", "w", stdout, "ctx=bin")))
         exit (vaxc$errno);                                           

   memset (&SystemData, 0, sizeof(SystemData));
   CollectSyiStatic (&SystemData);

   CgiHttpRefererPtr = GetCgiVar("HTTP_REFERER");
   CgiScriptNamePtr = GetCgiVar("SCRIPT_NAME");
   CgiQueryStringPtr = GetCgiVar("QUERY_STRING");

   /* derive the URL scheme and host from the HTTP referrer field */
   zptr = (sptr = RefererHost) + sizeof(RefererHost)-1;
   cptr = CgiHttpRefererPtr;
   while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
   if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
   if (*cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
   while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
   *sptr = '\0';

   if (isdigit(*CgiQueryStringPtr))
   {
      /********/
      /* data */
      /********/

      RefreshInterval = atoi(CgiQueryStringPtr);
      if (RefreshInterval <= 0 || RefreshInterval > 60)
         RefreshInterval = REFRESH_INTERVAL;

      MonDeSiData (&SystemData, RefreshInterval);
   }
   else
   {
      /********/
      /* page */
      /********/

      if (cptr = strstr (CgiQueryStringPtr, "across="))
         NodesAcross = atoi(cptr+7);
      if (NodesAcross < 0 || NodesAcross > 4) NodesAcross = 1;

      if (cptr = strstr (CgiQueryStringPtr, "iframe="))
      {
         zptr = (sptr = IframeId) + sizeof(IframeId)-1;
         for (cptr += 7;
              *cptr && *cptr != '&' && sptr < zptr;
              *sptr++ = *cptr++);
         *sptr = '\0';
      }

      if (cptr = strstr (CgiQueryStringPtr, "interval="))
         RefreshInterval = atoi(cptr+9);
      if (RefreshInterval < 0 || RefreshInterval > 60)
         RefreshInterval = REFRESH_INTERVAL;
                         
      if (NodeList = strstr (CgiQueryStringPtr, "nodes="))
      {
         for (sptr = (NodeList += 6); *sptr && *sptr != '&'; sptr++);
         *sptr = '\0';
         UrlDecode (NodeList);
         /* eliminate those pesky W*s <CR><LF>s */
         for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
            if (*(unsigned short*)cptr == '\r\n') cptr++;
         *sptr = '\0';
         /* compress white-space */
         for (cptr = sptr = NodeList; *cptr; *sptr++ = *cptr++)
            while (isspace(*cptr) && isspace(*(cptr+1))) cptr++;
         *sptr = '\0';
      }
      else
         NodeList = "";

      if (ispunct(*CgiQueryStringPtr))
         MonDeSiForm (NodeList, NodesAcross, RefreshInterval);
      else
      if (IframeId[0])
         MonDeSiNode (&SystemData, IframeId, RefreshInterval);
      else
      if (*NodeList)
         MonDeSiPage (NodeList, NodesAcross, RefreshInterval);
      else
         MonDeSiPage (SystemData.NodeName, 1, RefreshInterval);
   }

   exit (SS$_NORMAL);
} 

/*****************************************************************************/
/*
Return to previous process priority (in case ambit raise was successful).
*/

void ReturnPrio ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (!PrevPrio) return;

   status = sys$setpri (0, 0, PrevPrio, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$setpri() %%X%08.08X\n", status);
} 

/*****************************************************************************/
/*
Provide the client with the configuration setup form.
*/

void MonDeSiForm
(
char *NodeList,
int NodesAcross,
int RefreshInterval
)
{
   char  *cptr, *sptr;
   char  NodeUrl [256];

   /*********/
   /* begin */
   /*********/

   if (!*NodeList)
   {
      /* concoct a node list entry */
      sptr = GetCgiVar ("REQUEST_SCHEME");
      if (!*sptr)
      {
         sptr = GetCgiVar ("SERVER_PORT");
         if (atoi(sptr) == 443)
            sptr = "https:";
         else
            sptr = "http:";
      }
      cptr = GetCgiVar ("HTTP_HOST");
      sprintf (NodeList = NodeUrl, "%s//%s%s", sptr, cptr, CgiScriptNamePtr);
   }

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
\
<style type=\"text/css\">\n\
\
body { background-color:DarkSlateGray; margin:1em; \
font-family: arial, helvetica, tahoma; text-align:center; }\n\
input { border-width:1px; border-style:outset; }\n\
textarea,select { border-width:1px; border-style:inset; }\n\
.pageTitle { font-size:133%%; font-weight:bold; letter-spacing:1px; \
text-decoration:underline; color:white; }\n\
.subTitle { font-size:90%%; font-style:italic; letter-spacing:1px; \
color:white; }\n\
\
</style>\n\
\
<title>%s</title>\n\
</head>\n\
<body>\n\
<form method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
\
<span class=\"pageTitle\">Monitor de Sistema</span>\n\
<p style=\"margin-top:0.5em;\">\
<span class=\"subTitle\">~ System Monitor ~</span>\n\
\
<p><textarea id=\"nodes\" name=\"nodes\" rows=\"10\" cols=\"64\" \
scroll=\"auto\">\
%s</textarea>\n\
\
<p><select name=\"across\">\n\
<option value=\"1\"%s>1\n\
<option value=\"2\"%s>2\n\
<option value=\"3\"%s>3\n\
<option value=\"4\"%s>4\n\
</select>\
<span style=\"color:white;\">&nbsp;&nbsp;node(s) across window</span>\n\
\
<p><span style=\"color:white;\">Sampe rate&nbsp;&nbsp;</span>\
<select name=\"interval\">\n\
<option value=\"0\"%s>off\n\
<option value=\"1\"%s>1\n\
<option value=\"2\"%s>2\n\
<option value=\"5\"%s>5\n\
<option value=\"15\"%s>15\n\
</select>\
<span style=\"color:white;\">&nbsp;&nbsp;seconds</span>\n\
\
<p><input type=\"submit\" value=\" Monitor \">&nbsp;&nbsp;\n\
<input type=\"reset\" value=\" Reset \" \
onclick=\"return confirm(\'Are you sure?\')\">&nbsp;&nbsp;\n\
<input type=\"reset\" value=\" Clear \" \
onclick=\"if(confirm(\'Are you sure?\'))\
document.getElementById(\'nodes\').value=\'\';return false;\">\n\
\
</form>\n\
</body>\n\
</html>\n",
            SoftwareID,
            CgiScriptNamePtr,
            /* should be no HTML-forbidden characters in this! */
            NodeList,
            NodesAcross == 1 ? " selected" : "",
            NodesAcross == 2 ? " selected" : "",
            NodesAcross == 3 ? " selected" : "",
            NodesAcross == 4 ? " selected" : "",
            RefreshInterval ==  0 ? " selected" : "",
            RefreshInterval ==  1 ? " selected" : "",
            RefreshInterval ==  2 ? " selected" : "",
            RefreshInterval ==  5 ? " selected" : "",
            RefreshInterval == 15 ? " selected" : "");
} 

/*****************************************************************************/
/*
Provide the page that contains one or more monitor panels (IFRAMEs containing
their own HTML 'page' content).
*/

void MonDeSiPage
(
char *NodeList,
int NodesAcross,
int RefreshInterval
)
{
   int  NodeCount = 0;
   char  *aptr, *cptr, *sptr, *zptr;
   char  HostUrl [128];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiPage() |%s|\n", NodeList);

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
\
<style type=\"text/css\">\n\
\
body { background-color:DarkSlateGray; margin:1em; \
font-family: arial, helvetica, tahoma; text-align:center; }\n\
@media print { body { background-color:white; } }\n\
input { border-width:1px; border-style:outset; }\n\
@media print { input { visibility:hidden; } }\n\
.node_iframe { height:%dpx; width:%dpx; border:0; padding:0; margin:5px; }\n\
\
</style>\n\
\
<script language=\"JavaScript\">\n\
\
function mdsMsgChannel (msg)\n\
{\n\
   // message posting browsers (HTML5)\n\
   if (msg.data.substring(0,3) == \"mds\")\n\
   {\n\
      eval(msg.data);\n\
      return;\n\
   }\n\
   if (msg.data != \"\") alert(\"ERROR:\"+msg.data);\n\
}\n\
\
try { var mdsMessageListener = \
window.addEventListener(\"message\",mdsMsgChannel,false); } \
catch(err) { }\n\
\
function mdsBackChannel (forEval)\n\
{\n\
   // inter-frame kludge\n\
   if (forEval.substring(0,4) == \"#mds\")\n\
   {\n\
      eval(forEval.substr(1));\n\
      return;\n\
   }\n\
   if (forEval != \"\") alert(\"ERROR:\"+forEval.substr(1));\n\
}\n\
\
function mdsToggleHistory (iframeId,visible,barHeight,historyHeight)\n\
{\n\
   // remotely called via IFC from MonDeSiNode()\n\
   var oNodeFrame = document.getElementById(iframeId);\n\
   if (visible)\n\
      var heightNew = barHeight + historyHeight + 8;\n\
   else\n\
      var heightNew = barHeight + 11;\n\
   oNodeFrame.height = oNodeFrame.style.height = heightNew + \"px\";\n\
   if (visible) try { if (oNodeFrame.offsetTop > 100) \
window.scrollTo(0,oNodeFrame.offsetTop-10); } catch (err) { }\n\
}\n\
\
</script>\n\
<title>%s</title>\n\
</head>\n\
<body>\n",
            NODE_HEIGHT, NODE_WIDTH,
            SoftwareID);

   cptr = NodeList;
   while (*cptr)
   {
      while (*cptr && isspace(*cptr)) cptr++;
      zptr = (sptr = HostUrl) + sizeof(HostUrl)-1;
      /* parse to the end of the scheme */
      while (*cptr && !isspace(*cptr) && *cptr != '/' && sptr < zptr)
         *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/' && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /* if no domain name */
      if (!strchr (HostUrl, '.'))
      {
         /* same host, just use the script component */
         sptr = HostUrl;
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      else
      /* if provides a script component */
      if (*cptr && !isspace(*cptr))
      {
         if (*cptr && *cptr == '/' && sptr < zptr) *sptr++ = *cptr++;
         while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      }
      else
      {
         /* append a script component */
         for (aptr = CgiScriptNamePtr; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      *sptr = '\0';
      while (*cptr && !isspace(*cptr)) cptr++;
      while (*cptr && isspace(*cptr)) cptr++;

      NodeCount++,
      fprintf (stdout,
"<iframe id=\"node%d\" class=\"node_iframe\" \
frameborder=\"0\" scrolling=\"no\" \
src=\"%s?iframe=node%d&interval=%d\"></iframe>\n",
               NodeCount,
               HostUrl,
               NodeCount,
               RefreshInterval);
   }

   fprintf (stdout,
"<form method=\"GET\" action=\"%s\" \
enctype=\"application/x-www-form-urlencoded\" autocomplete=\"off\">\n\
<input type=\"hidden\" name=\"!\" value=\"\">\n\
<input type=\"hidden\" name=\"nodes\" value=\"%s\">\n\
<input type=\"hidden\" name=\"across\" value=\"%d\">\n\
<input type=\"hidden\" name=\"interval\" value=\"%d\">\n\
<input type=\"submit\" value=\" Configure \">\n\
</form>\n\
</body>\n\
</html>\n",
            CgiScriptNamePtr,
            strchr(NodeList,'.') ? NodeList : "",
            NodesAcross,
            RefreshInterval);
} 

/*****************************************************************************/
/*
Provide the in-frame node display.
*/

void MonDeSiNode
(
SYSTEM_DATA *sysptr,
char *IframeId,
int RefreshInterval
)
{
   char  DivStyle [256];

   /*********/
   /* begin */
   /*********/

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<html>\n\
<head>\n\
\
<style type=\"text/css\">\n\
\
body { background-color:white; margin:0; padding:0; \
font-family: arial, helvetica, tahoma; }\n\
.node_block { width:100%%; background-color:beige; color:black; \
border-style:outset; border-width:1px; border-color:black; \
margin:0; padding:0; text-align:left; }\n\
@media print { .node_block { background-color:white; \
border-style:solid; border-width:2px; } }\n\
.node_block td { margin:10px; padding:0; }\n\
.node_graph { border-width:0; margin:0; padding:0; }\n\
.node_graph td { margin:0; padding:0; }\n\
.node_notes { vertical-align:top; }\n\
.node_name { font-size:120%%; font-weight:bold; text-decoration:underline; }\n\
.node_text { font-size:80%%; }\n\
.node_notes input { font-size:80%%; border-style:outset; \
border-width:1px; border-color:gray; width:4.2em; \
padding:0; margin-right:0.4em; }\n\
@media print { .node_notes input { visibility:hidden; } }\n\
.history_graph { border-width:0; margin:0; padding:0 0 5px 0; }\n\
.history_graph td { margin:0; padding:0; }\n\
\
</style>\n\
<script language=\"JavaScript\">\n\
\
var mdsAlertSecs = %d;\n\
var mdsRefreshInterval = %d;\n\
var mdsRefererHost = \"%s\"\n\
var mdsScriptName = \"%s\";\n\
var mdsNodeName = \"%s\";\n\
var mdsCpuCount = %d;\n\
var mdsIframeId = \"%s\";\n\
var mdsBarWidth = %d;\n\
var mdsBarHeight = %d;\n\
var mdsHistWidth = %d;\n\
var mdsHistHeight = %d;\n\
var mdsHistVisible = false;\n\
var mdsBody;\n\
\
</script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
<script language=\"JavaScript\" src=\"%s%s\"></script>\n\
\
<title>%s</title>\n\
</head>\n\
<body>\n",
            ALERT_SECONDS,
            RefreshInterval,
            RefererHost,
            CgiScriptNamePtr,
            sysptr->NodeName,
            sysptr->ActiveCpuCount,
            IframeId,
            BAR_WIDTH, BAR_HEIGHT, HISTORY_WIDTH, HISTORY_HEIGHT,
            RefererHost, "/mondesi/-/js/mondesi.js",
            RefererHost, "/mondesi/-/js/wz_jsgraphics.js",
            SoftwareID);

   fprintf (stdout, "<table class=\"node_block\">\n<tr><td>\n");

   /**************/
   /* bar graphs */
   /**************/

   sprintf (DivStyle,
            "position:relative;height:%dpx;width:%dpx;margin-left:5px;",
            BAR_HEIGHT, BAR_WIDTH);

   fprintf (stdout,
"<table id=\"node_graph\" class=\"node_graph\">\n\
<tr><td><div id=\"node_notes\" \
style=\"width:%dpx;height:%dpx;margin:5px 5px 0 5px;\"></div></td>\n\
\
<td class=\"node_notes\" rowspan=\"2\">\n\
<div id=\"node_cpu\" style=\"%smargin-top:5px;\"></div>\n\
<div id=\"node_mem\" style=\"%s\"></div>\n\
<div id=\"node_pgfl\" style=\"%s\"></div>\n\
<div id=\"node_flts\" style=\"%s\"></div>\n\
<div id=\"node_bio\" style=\"%s\"></div>\n\
<div id=\"node_dio\" style=\"%s\"></div>\n\
<div id=\"node_lock\" style=\"%s\"></div>\n\
<div id=\"node_disk\" style=\"%s\"></div>\n\
<div id=\"node_mscp\" style=\"%s\"></div>\n\
<div id=\"node_debug\"></div>\n\
</td></tr>\n\
\
<tr><td class=\"node_notes\">\
<div style=\"width:%dpx;height:20px;margin:0 0 0 5px;white-space:nowrap;\">\n",
            HISTORY_WIDTH+2,
            (BAR_HEIGHT * 9) - 20,
            DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            HISTORY_WIDTH+2);

   fprintf (stdout,
"<input type=\"hidden\" name=\"refresh\" value=\"%d\">\n\
<input id=\"history_btn\" type=\"submit\" value=\"history\" \
onclick=\"mdsHistoryClick()\" style=\"width:5.0em;\">\
<input id=\"print_btn\" type=\"submit\" value=\"print\" \
onclick=\"mdsPrintNode();return false;\">\
<input type=\"submit\" value=\"scale\" \
onclick=\"return mdsScale();\">\
<input type=\"submit\" value=\"zero\" \
onclick=\"return mdsNodeZero();\">\
<input id=\"start_stop_btn\" type=\"submit\" value=\"stop\" \
onclick=\"return mdsStartStop();\">\n",
            RefreshInterval);

   fprintf (stdout,
"</div></td></tr>\n\
</table>\n\
</td></tr>\n\
<tr><td>\n");

   /******************/
   /* history graphs */
   /******************/

   sprintf (DivStyle,
            "position:relative;height:%dpx;width:%dpx;margin:5px;",
            HISTORY_HEIGHT+2, HISTORY_WIDTH+2);

   fprintf (stdout,
"<table id=\"history_graph\" class=\"history_graph\">\n\
\
<tr><td>\n\
<div id=\"hist_cpu\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_mem\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_pgfl\" style=\"%s\"></div>\n\
</td></tr>\n\
\
<tr><td>\n\
<div id=\"hist_flts\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_bio\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_dio\" style=\"%s\"></div>\n\
</td></tr>\n\
\
<tr><td>\n\
<div id=\"hist_lock\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_disk\" style=\"%s\"></div>\n\
</td><td>\n\
<div id=\"hist_mscp\" style=\"%s\"></div>\n\
</td></tr>\n\
\
</table>\n",
            DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle,
            DivStyle, DivStyle, DivStyle);

   fputs ("</td></tr>\n</table>\n", stdout);

   /**************/
   /* javascript */
   /**************/

   fprintf (stdout,
"<script language=\"JavaScript\">\n\
\
mdsBody = document.getElementsByTagName(\"body\").item(0);\n\
mdsInitialStart();\n\
mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\");\n\
\
mdsNodeInit(\"cpu\",true);\n\
mdsNodeInit(\"mem\",true);\n\
mdsNodeInit(\"pgfl\",true);\n\
mdsNodeInit(\"flts\",true);\n\
mdsNodeInit(\"bio\",true);\n\
mdsNodeInit(\"dio\",true);\n\
mdsNodeInit(\"lock\",true);\n\
mdsNodeInit(\"disk\",true);\n\
mdsNodeInit(\"mscp\",%s);\n\
\
</script>\n",
            sysptr->NodeName,
            sysptr->ClusterMember ? "true" : "false");

   fputs ("</body>\n</html>\n", stdout);
} 

/*****************************************************************************/
/*
At the defined interval gather dynamic system and performance data and for
appropriate (changed) data call the relevant JavaScript (RPC) function in the
parent panel (window) to process and graph associated data.  Write multiple
function calls into a dynamic buffer and then write that to <stdout> to
minimise the number of buffered I/O.  Do the best we can to disable proxy
caching ensuring a timely flow of response data.
*/

void MonDeSiData
(
SYSTEM_DATA *sysptr,
int RefreshInterval
)
{
   static $DESCRIPTOR (DateTimeFaoDsc, "!17%D\0");
   static $DESCRIPTOR (UpTimeFaoDsc, "!10%D\0");

   int  status,
        HostUrlLength,
        PrevBufferedIO = -1,
        PrevBusyCpu = -1,
        PrevClusterNodes = 0,
        PrevDirectIO = -1,
        PrevFcpRead = -1,
        PrevFcpWrite = -1,
        PrevLockActivity = -1,
        PrevMemInUse = -1,
        PrevMinute = -1,
        PrevMscpRead = -1,
        PrevMscpWrite = -1,
        PrevPageFaults = -1,
        PrevPageFileFree = -1,
        PrevProcessCount = -1,
        PrevUserCpu = -1,
        TimeStamp;
   unsigned long  UpBinTime [2];
   char  *cptr, *sptr, *zptr;
   char  CpuActive [32],
         DateTime [32],
         RpcBuffer [2048],
         Stupid_njobcnt [16],
         UpTime [32];
   $DESCRIPTOR (DateTimeDsc, DateTime);
   $DESCRIPTOR (UpTimeDsc, UpTime);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MonDeSiData()\n");

   /* derive the host URL from the HTTP referrer field */
   cptr = CgiHttpRefererPtr;
   while (*cptr && *(unsigned short*)cptr != '//') cptr++;
   if (*cptr) cptr += 2;
   while (*cptr && *cptr != '/') cptr++;
   HostUrlLength = cptr - CgiHttpRefererPtr;

   fprintf (stdout,
"Status: 200\n\
Content-Type: text/html\n\
Script-Control: X-stream-mode=1\n\
Script-Control: X-content-encoding-gzip=0\n\
Pragma: no-cache\n\
Cache-Control: no-cache, no-store, private\n\
\n\
<body onload=\"parent.mdsInitialStart();\">\n");
   fflush (stdout);

   for(;;)
   {
      /*****************/
      /* infinite loop */
      /*****************/

      sys$gettim (&CurrentBinTime);
      sys$numtim (&CurrentNumTime, &CurrentBinTime);
      TimeStamp = decc$fix_time (&CurrentBinTime);

      CollectSyiDynamic (sysptr);
      CollectRMI (sysptr);

      zptr = (sptr = RpcBuffer) + sizeof(RpcBuffer)-1;
      sptr += sprintf (sptr, "<script language=\"JavaScript\">");

      if (PrevMinute != CurrentNumTime[4] ||
          sysptr->ProcessCount != PrevProcessCount ||
          sysptr->ClusterNodes != PrevClusterNodes)
      {
         PrevMinute = CurrentNumTime[4];
         PrevProcessCount = sysptr->ProcessCount;

         sys$fao (&DateTimeFaoDsc, 0, &DateTimeDsc, &CurrentBinTime);
         lib$sub_times (&CurrentBinTime, &sysptr->BootBinTime, &UpBinTime);
         sys$fao (&UpTimeFaoDsc, 0, &UpTimeDsc, &UpBinTime);

         if (sysptr->ActiveCpuCount == sysptr->AvailCpuCount)
            sprintf (CpuActive, "%d", sysptr->ActiveCpuCount);
         else
            sprintf (CpuActive, "%d/%d", sysptr->ActiveCpuCount,
                                         sysptr->AvailCpuCount);

         sptr += sprintf (sptr,
"mdsCpuCount=%d;\
parent.mdsNodeNotes(\"<span class=\'node_name\'>%s</span>\
<span class=\'node_text\'>\
<br>%s<br>%s %s&nbsp;CPU%s %d&nbsp;MB %s&nbsp;%s&nbsp <nobr>Up %s</nobr>",
                          sysptr->ActiveCpuCount,
                          sysptr->NodeName,
                          DateTime,
                          sysptr->HwName,
                          CpuActive,
                          sysptr->AvailCpuCount > 1 ? "s" : "",
                          sysptr->MemSize,
                          sysptr->ArchName,
                          sysptr->VmsVersion,
                          UpTime);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         if (sysptr->ClusterMember)
         {
            sptr += sprintf (sptr,
"&nbsp Cluster nodes:%d votes:%d expect:%d quorum:%d",
                             sysptr->ClusterNodes,
                             sysptr->ClusterVotes,
                             sysptr->ClusterEVotes,
                             sysptr->ClusterQuorum);
            if (sptr > zptr) exit (SS$_BUGCHECK);
            PrevClusterNodes = sysptr->ClusterNodes;
         }

         if ((int)sysptr->ProcessNetCount >= 0)
            sprintf (Stupid_njobcnt, "%d", sysptr->ProcessNetCount);
         else
            strcpy (Stupid_njobcnt, "?");
         sptr += sprintf (sptr,
"&nbsp; Processes:%d interactive:%d batch:%d network:%s&nbsp; \
<span style=\'color:gray\'>(v%s/%s%s%s)</span></span>\");",
                          sysptr->ProcessCount,
                          sysptr->ProcessIntCount,
                          sysptr->ProcessBatCount,
                          Stupid_njobcnt,
                          SOFTWAREVN,
                          !ApacheHttpd && !OsuHttpd ? "WASD" : "",
                          ApacheHttpd ? "Apache" : "",
                          OsuHttpd ? "OSU" : "");
         if (sptr > zptr) exit (SS$_BUGCHECK);

         /* ensure memory in use and page file is refreshed */
         PrevMemInUse = PrevPageFileFree = -1;
      }

      if (sysptr->CpuBusyPercent != PrevBusyCpu ||
          sysptr->CpuUserPercent != PrevUserCpu)
      {
         char  CpuRaw [64];

         if (sysptr->CpuTicksRaw[1])
            /* bit of kludge to cater for potential quadword */
            sprintf (CpuRaw, "%d+(%d*4294967296)",
                             sysptr->CpuTicksRaw[0],
                             sysptr->CpuTicksRaw[1]);
         else
            sprintf (CpuRaw, "%d", sysptr->CpuTicksRaw[0]);

         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"cpu\",%d,%d,%d,%s,%u);",
                          sysptr->CpuBusyPercent,
                          sysptr->CpuBusyPercent - sysptr->CpuUserPercent,
                          sysptr->ActiveCpuCount*100,
                          CpuRaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevBusyCpu = sysptr->CpuBusyPercent;
         PrevUserCpu = sysptr->CpuUserPercent;
      }

      if (sysptr->MemInUse != PrevMemInUse)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"mem\",%d,-1,%d,-1,%u);",
                          sysptr->MemInUse,
                          sysptr->MemSize,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevMemInUse = sysptr->MemInUse;
      }

      if (sysptr->PageFileFree != PrevPageFileFree)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"pgfl\",%d,-1,%d,-1,%u);",
                          (sysptr->PageFilePage - sysptr->PageFileFree) >> 7,
                          sysptr->PageFilePage >> 7,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevPageFileFree = sysptr->PageFileFree;
      }

      if (sysptr->PageFaults != PrevPageFaults)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"flts\",%d,%d,-1,%d,%u);",
                          sysptr->PageFaultsSoft + sysptr->PageFaultsHard,
                          sysptr->PageFaultsHard,
                          sysptr->PageFaultsRaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevPageFaults = sysptr->PageFaults;
      }

      if (sysptr->BufferedIO != PrevBufferedIO)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"bio\",%d,-1,-1,%d,%u);",
                          sysptr->BufferedIO,
                          sysptr->BufferedIORaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevBufferedIO = sysptr->BufferedIO;
      }

      if (sysptr->DirectIO != PrevDirectIO)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"dio\",%d,-1,-1,%d,%u);",
                          sysptr->DirectIO,
                          sysptr->DirectIORaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevDirectIO = sysptr->DirectIO;
      }

      if (sysptr->LockActivity != PrevLockActivity)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"lock\",%d,-1,-1,%d,%u);",
                          sysptr->LockActivity,
                          sysptr->LockActivityRaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevLockActivity = sysptr->LockActivity;
      }

      if (sysptr->FcpRead != PrevFcpRead ||
          sysptr->FcpWrite != PrevFcpWrite)
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"disk\",%d,%d,-1,%d,%u);",
                          sysptr->FcpRead + sysptr->FcpWrite,
                          sysptr->FcpWrite,
                          sysptr->FcpReadRaw + sysptr->FcpWriteRaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevFcpRead = sysptr->FcpRead;
         PrevFcpWrite = sysptr->FcpWrite;
      }

      if (sysptr->ClusterMember &&
          (sysptr->MscpRead != PrevMscpRead ||
           sysptr->MscpWrite != PrevMscpWrite))
      {
         sptr += sprintf (sptr,
                          "parent.mdsMonVal(\"mscp\",%d,%d,-1,%d,%u);",
                          sysptr->MscpRead + sysptr->MscpWrite,
                          sysptr->MscpWrite,
                          sysptr->MscpReadRaw + sysptr->MscpWriteRaw,
                          TimeStamp);
         if (sptr > zptr) exit (SS$_BUGCHECK);

         PrevMscpRead = sysptr->MscpRead;
         PrevMscpWrite = sysptr->MscpWrite;
      }

      sptr += sprintf (sptr, "</script>\n");
      if (sptr > zptr) exit (SS$_BUGCHECK);

      if (PrevMinute != CurrentNumTime[4]) PrevMinute = CurrentNumTime[4];

      if (sptr > RpcBuffer+40)
      {
         fputs (RpcBuffer, stdout);
         fflush (stdout);
      }

      sleep (RefreshInterval);
   }

   /* should never ... */
} 

/*****************************************************************************/
/*
Per-boot system data.
*/ 

int CollectSyiStatic (SYSTEM_DATA *sysptr)

{
   static unsigned long  ActiveCpuCnt,
                         AvailCpuCount,
                         MemSize,
                         PageSize;
   static unsigned long  BootBinTime [2];
   static unsigned short  HwNameLength;
   static unsigned char  ClusterMember;
   static char  ArchName [15+1],
                HwName [1+60],
                NodeName [15+1],
                VmsVersion [8+1];

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList[] =
   {
       { sizeof(AvailCpuCount), SYI$_AVAILCPU_CNT, &AvailCpuCount, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(ArchName), SYI$_ARCH_NAME, &ArchName, 0 },
       { sizeof(BootBinTime), SYI$_BOOTTIME, &BootBinTime, 0 },
       { sizeof(ClusterMember), SYI$_CLUSTER_MEMBER, &ClusterMember, 0 },
       { sizeof(HwName), SYI$_HW_NAME, &HwName, &HwNameLength },
       { sizeof(MemSize), SYI$_MEMSIZE, &MemSize, 0 },
       { sizeof(NodeName), SYI$_SCSNODE, &NodeName, 0 },
       { sizeof(PageSize), SYI$_PAGE_SIZE, &PageSize, 0 },
       { sizeof(VmsVersion), SYI$_VERSION, &VmsVersion, 0 },
       {0,0,0,0}
   };

   int  status;
   char  *cptr, *sptr, *zptr;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiStatic()\n");

   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            &ItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.Status);

   if (VMSnok (status)) return (status);

   zptr = (sptr = sysptr->ArchName) + sizeof(sysptr->ArchName)-1;
   for (cptr = ArchName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   HwName[HwNameLength] = '\0';
   zptr = (sptr = sysptr->HwName) + sizeof(sysptr->HwName)-1;
   for (cptr = HwName; *cptr && sptr < zptr; *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = sysptr->NodeName) + sizeof(sysptr->NodeName)-1;
   for (cptr = NodeName;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   zptr = (sptr = sysptr->VmsVersion) + sizeof(sysptr->VmsVersion)-1;
   for (cptr = VmsVersion;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++); 
   *sptr = '\0';

   sysptr->ClusterMember = ClusterMember;
   /* convert to MBytes */
   sysptr->MemSize = MemSize >> 7;
   sysptr->PageSize = PageSize;

   sysptr->BootBinTime[0] = BootBinTime[0];
   sysptr->BootBinTime[1] = BootBinTime[1];

   sysptr->AvailCpuCount = AvailCpuCount;
   sysptr->ActiveCpuCount = ActiveCpuCnt;

   if (Debug)
      fprintf (stdout, "|%s|%s|%s|%s| %d %d\n",
               sysptr->ArchName, sysptr->HwName,
               sysptr->NodeName, sysptr->VmsVersion,
               sysptr->AvailCpuCount, ClusterMember);

   return (status);
} 

/*****************************************************************************/
/*
Dynamic system data.
*/ 

int CollectSyiDynamic (SYSTEM_DATA *sysptr)

{
   static unsigned long  ActiveCpuCnt,
                         PageFileFree,
                         PageFilePage;
   static unsigned short  ClusterEVotes,
                          ClusterNodes,
                          ClusterQuorum,
                          ClusterVotes;

   /* initialize structure for SYI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ClusterItemList[] =
   {
       { sizeof(ClusterNodes), SYI$_CLUSTER_NODES, &ClusterNodes, 0 },
       { sizeof(ClusterQuorum), SYI$_CLUSTER_QUORUM, &ClusterQuorum, 0 },
       { sizeof(ClusterVotes), SYI$_CLUSTER_VOTES, &ClusterVotes, 0 },
       { sizeof(ClusterEVotes), SYI$_CLUSTER_EVOTES, &ClusterEVotes, 0 },
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   }, NonItemList[] =
   {
       { sizeof(ActiveCpuCnt), SYI$_ACTIVECPU_CNT, &ActiveCpuCnt, 0 },
       { sizeof(PageFileFree), SYI$_PAGEFILE_FREE, &PageFileFree, 0 },
       { sizeof(PageFilePage), SYI$_PAGEFILE_PAGE, &PageFilePage, 0 },
       {0,0,0,0}
   };

   int  status;
   IOSBLK  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectSyiDynamic()\n");

   /* collect SYstem Information */
   status = sys$getsyiw (
            0,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            sysptr->ClusterMember ?
               (void*)&ClusterItemList :
               (void*)&NonItemList,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getsyi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.Status);

   if (VMSnok (status)) return (status);

   if (sysptr->ClusterMember)
   {
      sysptr->ClusterEVotes = ClusterEVotes;
      sysptr->ClusterNodes = ClusterNodes;
      sysptr->ClusterQuorum = ClusterQuorum;
      sysptr->ClusterVotes = ClusterVotes;
   }

   sysptr->ActiveCpuCount = ActiveCpuCnt;
   sysptr->PageFileFree = PageFileFree;
   sysptr->PageFilePage = PageFilePage;

   return (status);
} 

/*****************************************************************************/
/*
Collect System Performance Information.

Uses the system service $GETRMI() available for VMS V7.3 and later.

Cannot use RMI$_CPU* or RMI$_PROC*CNT on VMS V7.3-1 (and V7.3?) because they
don't seem to exist!  Fall back to handcrafting CPU from RMI$_MODES and
sys$gl_*jobcnt (many thanks to Graham Burley for the suggestion).  This
approach maintains two item lists and if the V7.3-2 (and later) fails with a
SS$$_BADPARAM switches to the other list.  Subsequent alternate processing is
performed by testing which list is in use.
*/ 

int CollectRMI (SYSTEM_DATA *sysptr)

{
/* required for VMS V7.3-1 and earlier */
#define INTSTK_MODE  0
#define MPSYNCH_MODE 1
#define KERNEL_MODE  2
#define EXEC_MODE    3
#define SUPER_MODE   4
#define USER_MODE    5
/* 6 is reserved */
#define IDLE_MODE    7

#define MODE_COUNT IDLE_MODE+1
#define CPU_MAX             32

#pragma member_alignment save 
#pragma nomember_alignment 

   /* structure for a single CPU's mode ticks */
   static struct SingleCPU {
       unsigned char  CPUid;
       unsigned long  Count [MODE_COUNT];
   };

   /* a structure all CPU's mode ticks for a system with up to 32 CPUs */
   static struct {
       unsigned long  NumberOfCPUs;
       struct SingleCPU  CPU [CPU_MAX];
   } ModeTicks;

#pragma member_alignment restore

   static const long cvtf_mode = LIB$K_DELTA_SECONDS_F;

   static unsigned long  BufferedIO,
                         DirectIO,
                         FcpRead,
                         FcpWrite,
                         LockConvert,
                         LockDequeue,
                         LockEnqueue,
                         LockWait,
                         MscpRead,
                         MscpWrite,
                         PageFaults,
                         PageReadIO,
                         PageWriteIO,
                         ProcessCount,
                         ProcessBatCount,
                         ProcessIntCount,
                         ProcessNetCount,
                         SpiEfn;

   static unsigned long  CpuExec [2],
                         CpuIntStk [2],
                         CpuKernel [2],
                         CpuMpSynch [2],
                         CpuSuper [2],
                         CpuUser [2];

#ifdef __ia64
   static unsigned long  FreeList [2],
                         ModList [2];
#else
   static unsigned long  FreeList,
                         ModList;
#endif

   static unsigned long  MscpEverything [35];

   /* initialize structure for RMI items */
   static struct {
      short  BufferLength;
      short  ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   } ItemList1[] =
   {
       /* for 7.3-2(?) and later which has the RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(CpuExec), RMI$_CPUEXEC, &CpuExec, 0 },
       { sizeof(CpuIntStk), RMI$_CPUINTSTK, &CpuIntStk, 0 },
       { sizeof(CpuKernel), RMI$_CPUKERNEL, &CpuKernel, 0 },
       { sizeof(CpuMpSynch), RMI$_CPUMPSYNCH, &CpuMpSynch, 0 },
       { sizeof(CpuSuper), RMI$_CPUSUPER, &CpuSuper, 0 },
       { sizeof(CpuUser), RMI$_CPUUSER, &CpuUser, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(MscpEverything), RMI$_MSCP_EVERYTHING, &MscpEverything, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       { sizeof(ProcessIntCount), RMI$_PROCINTCNT, &ProcessIntCount, 0 },
       { sizeof(ProcessBatCount), RMI$_PROCBATCNT, &ProcessBatCount, 0 },
       { sizeof(ProcessNetCount), RMI$_PROCNETCNT, &ProcessNetCount, 0 },
       {0,0,0,0}
   }, ItemList2[] =
   {
       /* for 7.3-1(?) and earlier which do not have RMI$_CPU_... items */
       { sizeof(BufferedIO), RMI$_BUFIO, &BufferedIO, 0 },
       { sizeof(ModeTicks), RMI$_MODES, &ModeTicks, 0 },
       { sizeof(LockDequeue), RMI$_DEQ, &LockDequeue, 0 },
       { sizeof(DirectIO), RMI$_DIRIO, &DirectIO, 0 },
       { sizeof(LockConvert), RMI$_ENQCVT, &LockConvert, 0 },
       { sizeof(LockEnqueue), RMI$_ENQNEW, &LockEnqueue, 0 },
       { sizeof(LockWait), RMI$_ENQWAIT, &LockWait, 0 },
       { sizeof(FcpRead), RMI$_FCPREAD, &FcpRead, 0 }, 
       { sizeof(FcpWrite), RMI$_FCPWRITE, &FcpWrite, 0 }, 
       { sizeof(FreeList), RMI$_FRLIST, &FreeList, 0 }, 
       { sizeof(ModList), RMI$_MODLIST, &ModList, 0 },
       { sizeof(MscpRead), RMI$_MSCP_READ, &MscpRead, 0 },
       { sizeof(MscpWrite), RMI$_MSCP_WRITE, &MscpWrite, 0 },
       { sizeof(PageFaults), RMI$_FAULTS, &PageFaults, 0 },
       { sizeof(PageReadIO), RMI$_PREADIO, &PageReadIO, 0 },
       { sizeof(PageWriteIO), RMI$_PWRITIO, &PageWriteIO, 0 },
       { sizeof(ProcessCount), RMI$_PROCS, &ProcessCount, 0 },
       {0,0,0,0}
   }, *ItemListPtr = ItemList1;

   int  cidx, midx, status, tlong;
   unsigned long  CpuTicksBusy [2],
                  DiffBinTime [2],
                  BinTime [2];
   float  tfloat,
          DeltaSeconds;
   IOSBLK  IOsb,
           SynchIOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CollectRMI()\n");

   if (!SpiEfn)
      if (VMSnok (status = lib$get_ef (&SpiEfn)))
         exit (status);

   /* get time immediately before getting system performance information */
   sys$gettim (&BinTime);

   /*
      If zero (first previous) was subtracted from the current time *some*
      Alpha lib$cvtf...() spat with a HPARITH (high performance arithmetic
      trap) FLTINV (floating invalid operation).  Not sure why!
   */
   if (!sysptr->PrevBinTime[0] && !sysptr->PrevBinTime[1])
   {
      sysptr->PrevBinTime[0] = BinTime[0];
      sysptr->PrevBinTime[1] = BinTime[1];
   }
   lib$sub_times (&BinTime, &sysptr->PrevBinTime, &DiffBinTime);
   sysptr->PrevBinTime[0] = BinTime[0];
   sysptr->PrevBinTime[1] = BinTime[1];
#ifdef __ia64
   lib$cvts_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#else
   lib$cvtf_from_internal_time (&cvtf_mode, &DeltaSeconds, &DiffBinTime);
#endif
   if (Debug) fprintf (stdout, "DeltaSeconds: %f\n", DeltaSeconds);

   /* collect System Performance Information */
   status = sys$getrmi (
            SpiEfn,  /* efn */
            0,  /* csiaddr */
            0,  /* nodename */
            ItemListPtr,  /* item list */
            &IOsb,  /* iosb */
            0,      /* astaddr */
            0 );    /* astprm */

   if (Debug)
      fprintf (stdout, "sys$getrmi() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, IOsb.Status);

   /* fall back to using the pre VMS V7.3-2 items */ 
   if (status == SS$_BADPARAM || IOsb.Status == SS$_BADPARAM)
   {
      if (ItemListPtr == ItemList1)
         ItemListPtr = ItemList2;
      else
      {
         fprintf (stdout, "<br>SS$_BADPARAM at line %d\n", __LINE__);
         return (status);
      }
   }

   if (VMSnok (status)) return (status);

/***
   status = sys$synch (SpiEfn, &SynchIOsb);

   if (Debug)
      fprintf (stdout, "sys$synch() %%X%08.08X  IOsb: %%X%08.08X\n",
               status, SynchIOsb.Status);

   if (VMSnok (status)) return (status);
***/

   /*******/
   /* CPU */
   /*******/

   if (ItemListPtr == ItemList1)
   {
      /* VMS 7.3-2 and later */
      status = lib$addx (&CpuExec, &CpuIntStk, &CpuTicksBusy, 0);
      if (VMSnok(status)) exit (status);
      status = lib$addx (&CpuKernel, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) exit (status);
      status = lib$addx (&CpuMpSynch, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) exit (status);
      status = lib$addx (&CpuSuper, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) exit (status);
      status = lib$addx (&CpuUser, &CpuTicksBusy, &CpuTicksBusy, 0);
      if (VMSnok(status)) exit (status);
   }
   else
   {
     /* VMS V7.3-1 and earlier */
     int  cidx, midx;
     for (cidx = 1; cidx < ModeTicks.NumberOfCPUs; cidx++)
        for (midx = INTSTK_MODE; midx <= IDLE_MODE; midx++)
            ModeTicks.CPU[0].Count[midx] += ModeTicks.CPU[cidx].Count[midx];

     CpuTicksBusy[1] = CpuUser[1] = 0;
     CpuTicksBusy[0] = ModeTicks.CPU[0].Count[INTSTK_MODE];
     /* interrupt item is actually the sum of interrupt and idle! */
     CpuTicksBusy[0] -= ModeTicks.CPU[0].Count[IDLE_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[KERNEL_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[MPSYNCH_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[EXEC_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[SUPER_MODE];
     CpuTicksBusy[0] += ModeTicks.CPU[0].Count[USER_MODE];
     CpuUser[0] = ModeTicks.CPU[0].Count[USER_MODE];

#ifdef  __VAX
     ProcessIntCount = sys$gw_ijobcnt;
     ProcessBatCount = sys$gw_bjobcnt;
     ProcessNetCount = sys$gw_njobcnt;
#else
     ProcessIntCount = sys$gl_ijobcnt;
     ProcessBatCount = sys$gl_bjobcnt;
     ProcessNetCount = sys$gl_njobcnt;
#endif
     /* of course sys$g*_njobcnt doesn't seem to account network jobs! */
     ProcessNetCount = -1;
   }

   if (sysptr->PrevCpuTicksBusy[0] || sysptr->PrevCpuTicksBusy[1])
   {
      status = lib$subx (&CpuTicksBusy, &sysptr->PrevCpuTicksBusy,
                         &sysptr->CpuTicksBusy, 0);
      if (VMSnok(status)) exit (status);

      tfloat = ((float)sysptr->CpuTicksBusy[0] +
                ((float)sysptr->CpuTicksBusy[1] * 4294967296.0));
      tfloat /= DeltaSeconds;
      sysptr->CpuBusyPercent = (unsigned int)tfloat;
      if (sysptr->CpuBusyPercent > 100 * sysptr->ActiveCpuCount)
         sysptr->CpuBusyPercent = 100 * sysptr->ActiveCpuCount;
   }
   else
      sysptr->CpuBusyPercent = 0;

   sysptr->PrevCpuTicksBusy[0] = sysptr->CpuTicksRaw[0] = CpuTicksBusy[0];
   sysptr->PrevCpuTicksBusy[1] = sysptr->CpuTicksRaw[1] = CpuTicksBusy[1];

   if (sysptr->PrevCpuTicksUser[0] || sysptr->PrevCpuTicksUser[1])
   {
      status = lib$subx (&CpuUser, &sysptr->PrevCpuTicksUser,
                         &sysptr->CpuTicksUser, 0);
      if (VMSnok(status)) exit (status);

      tfloat = ((float)sysptr->CpuTicksUser[0] +
                ((float)sysptr->CpuTicksUser[1] * 4294967296.0));
      tfloat /= DeltaSeconds;
      sysptr->CpuUserPercent = (unsigned int)tfloat;
      if (sysptr->CpuUserPercent > 100 * sysptr->ActiveCpuCount)
         sysptr->CpuUserPercent = 100 * sysptr->ActiveCpuCount;
   }
   else
      sysptr->CpuUserPercent = 0;

   sysptr->PrevCpuTicksUser[0] = CpuUser[0];
   sysptr->PrevCpuTicksUser[1] = CpuUser[1];

   if (Debug)
      fprintf (stdout, "CPU: %d%% USER: %d%%\n",
               sysptr->CpuBusyPercent, sysptr->CpuUserPercent);

   /***************/
   /* page faults */
   /***************/

   if (sysptr->PrevPageFaults)
      sysptr->PageFaults = PageFaults - sysptr->PrevPageFaults;
   else
      sysptr->PageFaults = 0;
   sysptr->PageFaultsRaw = sysptr->PrevPageFaults = PageFaults;

   if (sysptr->PrevPageReadIO)
      sysptr->PageReadIO = PageReadIO - sysptr->PrevPageReadIO;
   else
      sysptr->PageReadIO = 0;
   sysptr->PrevPageReadIO = PageReadIO;

   if (sysptr->PrevPageWriteIO)
      sysptr->PageWriteIO = PageWriteIO - sysptr->PrevPageWriteIO;
   else
      sysptr->PageWriteIO = 0;
   sysptr->PrevPageWriteIO = PageWriteIO;

   sysptr->PageFaultsHard = sysptr->PageReadIO + sysptr->PageWriteIO;
   tfloat = (float)sysptr->PageFaultsHard / DeltaSeconds;
   sysptr->PageFaultsHard = (unsigned int)tfloat;

   sysptr->PageFaultsSoft = sysptr->PageFaults - sysptr->PageFaultsHard;
   tfloat = (float)sysptr->PageFaultsSoft / DeltaSeconds;
   sysptr->PageFaultsSoft = (unsigned int)tfloat;

   /********/
   /* MSCP */
   /********/

   if (sysptr->ClusterMember)
   {
      if (ItemListPtr == ItemList1)
      {
         MscpRead = MscpEverything[23];
         MscpWrite = MscpEverything[24];
      }

      if (sysptr->PrevMscpRead)
         sysptr->MscpRead = MscpRead - sysptr->PrevMscpRead;
      else
         sysptr->MscpRead = 0;
      sysptr->MscpReadRaw = sysptr->PrevMscpRead = MscpRead;
      tfloat = (float)sysptr->MscpRead / DeltaSeconds;
      sysptr->MscpRead = (unsigned int)tfloat;

      if (sysptr->PrevMscpWrite)
         sysptr->MscpWrite = MscpWrite - sysptr->PrevMscpWrite;
      else
         sysptr->MscpWrite = 0;
      sysptr->MscpWriteRaw = sysptr->PrevMscpWrite = MscpWrite;
      tfloat = (float)sysptr->MscpWrite / DeltaSeconds;
      sysptr->MscpWrite = (unsigned int)tfloat;
   }

   /********/
   /* lock */
   /********/

   if (sysptr->PrevLockActivity)
      sysptr->LockActivity = LockEnqueue + LockDequeue + LockWait -
                             sysptr->PrevLockActivity;
   else
      sysptr->LockActivity = 0;
   sysptr->LockActivityRaw = sysptr->PrevLockActivity =
      LockEnqueue + LockDequeue + LockWait;
   tfloat = (float)sysptr->LockActivity / DeltaSeconds;
   sysptr->LockActivity = (unsigned int)tfloat;

   if (sysptr->PrevLockWait)
      sysptr->LockWait = LockWait - sysptr->PrevLockWait;
   else
      sysptr->LockWait = 0;
   sysptr->PrevLockWait = LockWait;

   /********/
   /* disk */
   /********/

   if (sysptr->PrevFcpRead)
      sysptr->FcpRead = FcpRead - sysptr->PrevFcpRead;
   else
      sysptr->FcpRead = 0;
   sysptr->FcpReadRaw = sysptr->PrevFcpRead = FcpRead;
   tfloat = (float)sysptr->FcpRead / DeltaSeconds;
   sysptr->FcpRead = (unsigned int)tfloat;

   if (sysptr->PrevFcpWrite)
      sysptr->FcpWrite = FcpWrite - sysptr->PrevFcpWrite;
   else
      sysptr->FcpWrite = 0;
   sysptr->FcpWriteRaw = sysptr->PrevFcpWrite = FcpWrite;
   tfloat = (float)sysptr->FcpWrite / DeltaSeconds;
   sysptr->FcpWrite = (unsigned int)tfloat;

   /*********/
   /* other */
   /*********/

   if (sysptr->PrevDirectIO)
      sysptr->DirectIO = DirectIO - sysptr->PrevDirectIO;
   else
      sysptr->DirectIO = 0;
   sysptr->DirectIORaw = sysptr->PrevDirectIO = DirectIO;
   tfloat = (float)sysptr->DirectIO / DeltaSeconds;
   sysptr->DirectIO = (unsigned int)tfloat;

   if (sysptr->PrevBufferedIO)
      sysptr->BufferedIO = BufferedIO - sysptr->PrevBufferedIO;
   else
      sysptr->BufferedIO = 0;
   sysptr->BufferedIORaw = sysptr->PrevBufferedIO = BufferedIO;
   tfloat = (float)sysptr->BufferedIO / DeltaSeconds;
   sysptr->BufferedIO = (unsigned int)tfloat;

#ifdef __ia64
   sysptr->MemInUse = sysptr->MemSize - (FreeList[0] >> 7) - (ModList[0] >> 7);
#else
   sysptr->MemInUse = sysptr->MemSize - (FreeList >> 7) - (ModList >> 7);
#endif
   if (sysptr->MemInUse > sysptr->MemInUseMax)
      sysptr->MemInUseMax = sysptr->MemInUse;

   sysptr->ProcessCount = ProcessCount;
   sysptr->ProcessBatCount = ProcessBatCount;
   sysptr->ProcessIntCount = ProcessIntCount;
   sysptr->ProcessNetCount = ProcessNetCount;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!  Returns the size of the resultant string.
*/

int UrlDecode (char *String)

{
   unsigned char  ch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            ch = 0;
            if (*cptr >= '0' && *cptr <= '9')
               ch = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            while (*cptr == '\r' || *cptr == '\n') cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               ch += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               ch += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = ch;

            continue;

         case '\r' :
         case '\n' :

            cptr++;
            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

/*****************************************************************************/
/*
Get CGI variables.  Return a pointer to the appropriate string.
*/

char* GetCgiVarWith (char*, char*);
char* GetCgiVar (char* VarName) { return GetCgiVarWith (VarName, ""); }
char* GetCgiVarNull (char* VarName) { return GetCgiVarWith (VarName, NULL); }

char* GetCgiVarWith
(
char *VarName,
char *NotFound
)
{
   static int  PrefixLength = -1;
   static char  VarNameBuf [128] = "WWW_";

   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (PrefixLength < 0)
   {
      /* adjust for variable Web server CGI variable prefix */
      if (getenv ("WWW_SERVER_SOFTWARE"))
         PrefixLength = 0;
      else
         PrefixLength = 4;
   }

   zptr = (sptr = VarNameBuf+4) + sizeof(VarNameBuf)-5-1;
   for (cptr = VarName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr = '\0';
   cptr = getenv (VarNameBuf + PrefixLength);
   if (Debug)
      fprintf (stdout, "GetCgiVarWith() |%s|%s|\n",
               VarNameBuf+PrefixLength, cptr ? cptr : "(null)");
   if (cptr) return (cptr);
   return (NotFound);
} 

/****************************************************************************/

